@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.13 2006/08/02 08:27:29 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool/main.c,v 1.21 2007/07/09 05:12:21 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2006	PgPool Global Development Group
+ * Copyright (c) 2003-2007	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -21,11 +21,15 @@
 #include "pool.h"
 
 #include <sys/types.h>
+#include <sys/time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <sys/un.h>
 #include <netdb.h>
 #include <arpa/inet.h>
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 
 #include <sys/stat.h>
 #include <fcntl.h>
@@ -46,6 +50,20 @@
 
 #include "version.h"
 
+#define CHECK_REQUEST \
+	do \
+	{ \
+		if (failover_request) \
+		{ \
+			failover(failover_request); \
+			failover_request = 0; \
+		} \
+		if (sigchld_request) \
+		{ \
+			reaper(); \
+		} \
+	} while (0)
+
 #define PGPOOLMAXLITSENQUEUELENGTH 10000
 static void daemonize(void);
 static int read_pid_file(void);
@@ -54,6 +72,10 @@ static pid_t fork_a_child(int unix_fd, int inet_fd);
 static int create_unix_domain_socket(void);
 static int create_inet_domain_socket(const char *hostname);
 static void myexit(int code);
+static void failover(int sig);
+static void reaper(void);
+static int pool_pause(struct timeval *timeout);
+static void pool_sleep(unsigned int second);
 
 static RETSIGTYPE exit_handler(int sig);
 static RETSIGTYPE reap_handler(int sig);
@@ -86,6 +108,12 @@ static int stop_sig = SIGTERM;	/* stopping signal default value */
 static int switch_over_sig = SIGUSR1;	/* switch over signal default value */
 
 static int health_check_timer_expired;		/* non 0 if health check timer expired */
+static volatile sig_atomic_t failover_request = 0;
+static volatile sig_atomic_t sigchld_request = 0;
+static int pipe_fds[2]; /* for delivering signals */
+
+int myargc;
+char **myargv;
 
 /*
 * pgpool main program
@@ -94,15 +122,29 @@ int main(int argc, char **argv)
 {
 	int opt;
 	char conf_file[POOLMAXPATHLEN+1];
+	char hba_file[POOLMAXPATHLEN+1];
 	int i;
 	int pid;
 
-	snprintf(conf_file, sizeof(conf_file), "%s/%s", DEFAULT_CONFIGDIR, POOL_CONF_FILE_NAME);
+	myargc = argc;
+	myargv = argv;
 
-	while ((opt = getopt(argc, argv, "df:hm:ns:")) != -1)
+	snprintf(conf_file, sizeof(conf_file), "%s/%s", DEFAULT_CONFIGDIR, POOL_CONF_FILE_NAME);
+	snprintf(hba_file, sizeof(hba_file), "%s/%s", DEFAULT_CONFIGDIR, HBA_CONF_FILE_NAME);
+
+	while ((opt = getopt(argc, argv, "a:df:hm:ns:")) != -1)
 	{
 		switch (opt)
 		{
+			case 'a':    /* specify hba configuration file */
+				if (!optarg)
+				{
+					usage();
+					exit(1);
+				}
+				strncpy(hba_file, optarg, sizeof(hba_file));
+				break;
+
 			case 'd':	/* debug option */
 				debug = 1;
 				break;
@@ -186,6 +228,10 @@ int main(int argc, char **argv)
 
 	pool_debug("weight: %ld", weight_master);
 
+	/* read pool_hba.conf */
+	if (pool_config.enable_pool_hba)
+		load_hba(hba_file);
+
 	/*
 	 * if a non-switch argument remains, then it should be either "stop" or "switch"
 	 */
@@ -275,6 +321,7 @@ int main(int argc, char **argv)
 	}
 
 	/* set up signal handlers */
+	POOL_SETMASK(&BlockSig);
 	pool_signal(SIGTERM, exit_handler);
 	pool_signal(SIGINT, exit_handler);
 	pool_signal(SIGQUIT, exit_handler);
@@ -283,6 +330,13 @@ int main(int argc, char **argv)
 	pool_signal(SIGUSR2, failover_handler);
 	pool_signal(SIGHUP, exit_handler);
 
+	/* create pipe for delivering event */
+	if (pipe(pipe_fds) < 0)
+	{
+		pool_error("failed to create pipe");
+		myexit(1);
+	}
+
 	pool_log("pgpool successfully started");
 
 	/*
@@ -290,6 +344,8 @@ int main(int argc, char **argv)
 	 */
 	for (;;)
 	{
+		CHECK_REQUEST;
+
 		/* do we need health checking for PostgreSQL? */
 		if (!degenerated && pool_config.health_check_period > 0)
 		{
@@ -320,11 +376,11 @@ int main(int argc, char **argv)
 			{
 				if (sts == -1)
 				{
-					failover_handler(SIGUSR1);		/* master down */
+					failover(SIGUSR1);		/* master down */
 				}
 				else if (sts == -2)
 				{
-					failover_handler(SIGUSR2);		/* secondary down */
+					failover(SIGUSR2);		/* secondary down */
 				}
 			}
 
@@ -335,14 +391,20 @@ int main(int argc, char **argv)
 			}
 
 			sleep_time = pool_config.health_check_period;
-			while (sleep_time > 0)
-			{
-				sleep_time = sleep(sleep_time);
-			}
+			pool_sleep(sleep_time);
 		}
 		else
 		{
-			pause();
+			for (;;)
+			{
+				int r;
+
+				POOL_SETMASK(&UnBlockSig);
+				r = pool_pause(NULL);
+				POOL_SETMASK(&BlockSig);
+				if (r > 0)
+					break;
+			}
 		}
 	}
 	return 0;
@@ -352,11 +414,12 @@ static void usage(void)
 {
 	fprintf(stderr, "pgpool version %s(%s),\n",	VERSION, PGPOOLVERSION);
 	fprintf(stderr, "  a generic connection pool/replication/load balance server for PostgreSQL\n\n");
-	fprintf(stderr, "usage: pgpool [-f config_file][-n][-d]\n");
-	fprintf(stderr, "usage: pgpool [-f config_file] [-m {s[mart]|f[ast]|i[mmediate]}] stop\n");
-	fprintf(stderr, "usage: pgpool [-f config_file] [-s {m[aster]|s[econdary]] switch\n");
+	fprintf(stderr, "usage: pgpool [-f config_file][-a hba_file][-n][-d]\n");
+	fprintf(stderr, "usage: pgpool [-f config_file][-a hba_file] [-m {s[mart]|f[ast]|i[mmediate]}] stop\n");
+	fprintf(stderr, "usage: pgpool [-f config_file][-a hba_file] [-s {m[aster]|s[econdary]] switch\n");
 	fprintf(stderr, "usage: pgpool -h\n");
 	fprintf(stderr, "  config_file default path: %s/%s\n",DEFAULT_CONFIGDIR, POOL_CONF_FILE_NAME);
+	fprintf(stderr, "  hba_file default path:    %s/%s\n",DEFAULT_CONFIGDIR, HBA_CONF_FILE_NAME);
 	fprintf(stderr, "  -n: don't run in daemon mode. does not detatch control tty\n");
 	fprintf(stderr, "  -d: debug mode. lots of debug information will be printed\n");
 	fprintf(stderr, "  stop: stop pgpool\n");
@@ -512,7 +575,7 @@ static void write_pid_file(void)
 /*
 * fork a child
 */
-pid_t fork_a_child(int unix_fd, int inet_fd)
+static pid_t fork_a_child(int unix_fd, int inet_fd)
 {
 	pid_t pid;
 
@@ -520,6 +583,8 @@ pid_t fork_a_child(int unix_fd, int inet_fd)
 
 	if (pid == 0)
 	{
+		myargv = save_ps_display_args(myargc, myargv);
+
 		/* call child main */
 		POOL_SETMASK(&UnBlockSig);
 		do_child(unix_fd, inet_fd);
@@ -729,7 +794,6 @@ static RETSIGTYPE exit_handler(int sig)
 	myexit(0);
 }
 
-
 /*
  * handle SIGUSR1/SIGUSR2 (backend connection error, fail over request, if possible)
  *
@@ -737,12 +801,22 @@ static RETSIGTYPE exit_handler(int sig)
  * if sig == SIGUSR2, we assume that the secondary has been down.
  */
 static RETSIGTYPE failover_handler(int sig)
+{
+	POOL_SETMASK(&BlockSig);
+	failover_request = 1;
+	write(pipe_fds[1], "\0", 1);
+	POOL_SETMASK(&UnBlockSig);
+}
+
+/*
+ * Process failover request
+ * failover() must be called under protecting signals.
+ */
+static void failover(int sig)
 {
 	int i;
 	int replication = 0;
 
-	POOL_SETMASK(&BlockSig);
-
 	pool_debug("failover_handler called");
 
 	/*
@@ -752,7 +826,6 @@ static RETSIGTYPE failover_handler(int sig)
 	if (getpid() != mypid)
 	{
 		pool_debug("failover_handler: I am not parent");
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
@@ -761,7 +834,6 @@ static RETSIGTYPE failover_handler(int sig)
 	 */
 	if (exiting)
 	{
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
@@ -770,7 +842,6 @@ static RETSIGTYPE failover_handler(int sig)
 	 */
 	if (switching)
 	{
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
@@ -934,7 +1005,6 @@ static RETSIGTYPE failover_handler(int sig)
 
 		switching = 0;
 	}
-	POOL_SETMASK(&UnBlockSig);
 }
 
 /*
@@ -951,24 +1021,33 @@ static RETSIGTYPE health_check_timer_handler(int sig)
  * handle SIGCHLD
  */
 static RETSIGTYPE reap_handler(int sig)
+{
+	POOL_SETMASK(&BlockSig);
+	sigchld_request = 1;
+	write(pipe_fds[1], "\0", 1);
+	POOL_SETMASK(&UnBlockSig);
+}
+
+/*
+ * Attach zombie processes and restart child processes.
+ * reaper() must be called under protecting signals.
+ */
+static void reaper(void)
 {
 	pid_t pid;
 	int status;
 	int i;
 
-	POOL_SETMASK(&BlockSig);
-
 	pool_debug("reap_handler called");
+	sigchld_request = 0;
 
 	if (exiting)
 	{
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
 	if (switching)
 	{
-		POOL_SETMASK(&UnBlockSig);
 		return;
 	}
 
@@ -997,6 +1076,59 @@ static RETSIGTYPE reap_handler(int sig)
 				}
 			}
 		}
-		POOL_SETMASK(&UnBlockSig);
 
 	}
+
+/*
+ * pool_pause: A process pauses by select().
+ */
+static int pool_pause(struct timeval *timeout)
+{
+	fd_set rfds;
+	int n;
+	char dummy;
+
+	FD_ZERO(&rfds);
+	FD_SET(pipe_fds[0], &rfds);
+	n = select(pipe_fds[0]+1, &rfds, NULL, NULL, timeout);
+	if (n == 1)
+		read(pipe_fds[0], &dummy, 1);
+	return n;
+}
+
+/*
+ * pool_pause: A process sleep using pool_pause().
+ *             If a signal event occurs, it raises signal handler.
+ */
+static void pool_sleep(unsigned int second)
+{
+	struct timeval current_time, sleep_time;
+
+	gettimeofday(&current_time, NULL);
+	sleep_time.tv_sec = second + current_time.tv_sec;
+	sleep_time.tv_usec = current_time.tv_usec;
+
+	POOL_SETMASK(&UnBlockSig);
+	while (sleep_time.tv_sec > current_time.tv_sec ||
+		   sleep_time.tv_usec > current_time.tv_usec)
+	{
+		struct timeval timeout;
+		int r;
+
+		timeout.tv_sec = sleep_time.tv_sec - current_time.tv_sec;
+		timeout.tv_usec = sleep_time.tv_usec - current_time.tv_usec;
+		if (timeout.tv_usec < 0)
+		{
+			timeout.tv_sec--;
+			timeout.tv_usec += 1000000;
+		}
+
+		r = pool_pause(&timeout);
+		POOL_SETMASK(&BlockSig);
+		if (r > 0)									
+			CHECK_REQUEST;
+		POOL_SETMASK(&UnBlockSig);
+		gettimeofday(&current_time, NULL);
+	}
+	POOL_SETMASK(&BlockSig);
+}
