@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool/pool_process_query.c,v 1.45 2007/05/17 02:05:50 yamaguti Exp $
+ * $Header: /cvsroot/pgpool/pgpool/pool_process_query.c,v 1.59 2007/08/01 03:54:56 y-asaba Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -41,6 +41,9 @@
 
 #define INIT_STATEMENT_LIST_SIZE 8
 
+#define DEADLOCK_ERROR_CODE "40P01"
+#define POOL_ERROR_QUERY "send invalid query from pgpool to abort transaction"
+
 typedef struct {
 	char *statement_name;
 	char *portal_name;
@@ -121,6 +124,7 @@ static int load_balance_enabled(POOL_CONNECTION_POOL *backend, char *sql);
 static void start_load_balance(POOL_CONNECTION_POOL *backend);
 static void end_load_balance(POOL_CONNECTION_POOL *backend);
 static POOL_STATUS do_command(POOL_CONNECTION *backend, char *query, int protoMajor, int no_ready_for_query);
+static POOL_STATUS do_error_command(POOL_CONNECTION *backend, int protoMajor);
 static int need_insert_lock(POOL_CONNECTION_POOL *backend, char *query);
 static POOL_STATUS insert_lock(POOL_CONNECTION_POOL *backend, char *query);
 static char *get_insert_command_table_name(char *query);
@@ -136,6 +140,7 @@ static PreparedStatement *lookup_prepared_statement_by_statement(PreparedStateme
 static PreparedStatement *lookup_prepared_statement_by_portal(PreparedStatementList *p, const char *name);
 static int send_deallocate(POOL_CONNECTION_POOL *backend, PreparedStatementList *p, int n);
 static char *normalize_prepared_stmt_name(const char *name);
+static POOL_STATUS error_kind_mismatch(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend, int kind, int kind1);
 
 static POOL_CONNECTION_POOL_SLOT *slots[MAX_CONNECTION_SLOTS];
 
@@ -152,8 +157,12 @@ static PreparedStatement *pending_prepared_stmt = NULL;
 static PreparedStatementList prepared_list; /* prepared statement name list */
 static PreparedStatement *unnamed_statement = NULL;
 static PreparedStatement *unnamed_portal = NULL;
+static int force_replication = 0; /* non 0 if force to replicate query */
+static int prepare_in_session = 0;
+
 static int is_drop_database(char *query);		/* returns non 0 if this is a DROP DATABASE command */
 static void query_ps_status(char *query, POOL_CONNECTION_POOL *backend);		/* show ps status */
+static int detect_deadlock_error(POOL_CONNECTION *master, int major);
 
 POOL_STATUS pool_process_query(POOL_CONNECTION *frontend, 
 							   POOL_CONNECTION_POOL *backend,
@@ -213,6 +222,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 			else	/* no more query(st == 2) */
 			{
 				frontend->no_forward = 0;
+				prepare_in_session = 0;
 				return POOL_CONTINUE;
 			}
 
@@ -321,31 +331,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 					pool_read(SECONDARY(backend), &kind1, 1);
 					if (kind == '\0' || kind != kind1)
 					{
-						int sts;
-
-						pool_error("pool_process_query: kind does not match between backends master(%c) secondary(%c)",
-								   kind, kind1);
-						pool_send_error_message(frontend, MAJOR(backend), "XX000", 
-												"kind mismatch between backends", "",
-												"check data consistency between master and secondary", __FILE__, __LINE__);
-
-						/* health check */
-						sts = health_check();
-						if (sts == -1)
-						{
-							notice_backend_error(1);
-							exit(1);
-						}
-						else if (sts == -2)
-						{
-							notice_backend_error(0);
-							exit(1);
-						}
-
-						if (pool_config.replication_stop_on_mismatch)
-							return POOL_FATAL;
-						else
-							return POOL_ERROR;
+						return error_kind_mismatch(frontend, backend, kind, kind1);
 					}
 				}
 				pool_debug("read kind from backend pending data %c len: %d po: %d", kind, MASTER(backend)->len, MASTER(backend)->po);
@@ -406,31 +392,7 @@ POOL_STATUS pool_process_query(POOL_CONNECTION *frontend,
 
 			if (kind == '\0' || kind != kind1)
 			{
-				int sts;
-
-				pool_error("pool_process_query: kind does not match between backends master(%c) secondary(%c)",
-						   kind, kind1);
-				pool_send_error_message(frontend, MAJOR(backend), "XX000", 
-										"kind mismatch between backends", "",
-										"check data consistency between master and secondary", __FILE__, __LINE__);
-
-				/* health check */
-				sts = health_check();
-				if (sts == -1)
-				{
-					notice_backend_error(1);
-					exit(1);
-				}
-				else if (sts == -2)
-				{
-					notice_backend_error(0);
-					exit(1);
-				}
-
-				if (pool_config.replication_stop_on_mismatch)
-					return POOL_FATAL;
-				else
-					return POOL_ERROR;
+				return error_kind_mismatch(frontend, backend, kind, kind1);
 			}
 		}
 
@@ -566,6 +528,7 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 	int len;
 	static char *sq = "show pool_status";
 	POOL_STATUS status;
+	int deadlock_detected = 0;
 
 	if (query == NULL)	/* need to read query from frontend? */
 	{
@@ -653,6 +616,9 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 		{
 			pending_function = add_prepared_list;
 			pending_prepared_stmt = stmt;
+			if (REPLICATION)
+				force_replication = 1;
+			prepare_in_session = 1;
 		}
 	}
 	else if (frontend &&
@@ -689,6 +655,8 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 			return POOL_END;
 		}
 		free(buf);
+		if (REPLICATION)
+			force_replication = 1;
 	}
 
 	if (frontend &&
@@ -717,16 +685,25 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 		string1 = string;
 	}
 
+	/* reset query have to be replicated */
+	if (frontend == NULL && prepare_in_session && REPLICATION)
+		force_replication = 1;
+
 	/* load balance trick */
 	if (load_balance_enabled(backend, string1))
 		start_load_balance(backend);
+	else if (force_replication)
+	{
+		replication_was_enabled = REPLICATION;
+		REPLICATION = 1;
+	}
 	else if (MASTER_SLAVE)
 	{
 		master_slave_was_enabled = 1;
 		MASTER_SLAVE = 0;
 		master_slave_dml = 1;
 	}
-	else if (REPLICATION && is_select_query(string1) && !is_sequence_query(string1))
+	else if (REPLICATION && !pool_config.replicate_select && is_select_query(string1) && !is_sequence_query(string1))
 	{
 		int i;
 		/* save backend connection slots */
@@ -794,6 +771,21 @@ static POOL_STATUS Query(POOL_CONNECTION *frontend,
 			pool_debug("waiting for master completing the query");
 			if (synchronize(MASTER(backend)))
 				return POOL_END;
+
+			/*
+			 * We must check deadlock error because a aborted transaction
+			 * by detecting deadlock isn't same on all nodes.
+			 * If a transaction is aborted on master node, pgpool send a
+			 * error query to another nodes.
+			 */
+			deadlock_detected = detect_deadlock_error(MASTER(backend), MAJOR(backend));
+			if (deadlock_detected < 0)
+				return POOL_END;
+			else if (deadlock_detected)
+			{
+				string = POOL_ERROR_QUERY;
+				len = strlen(string) + 1;
+			}
 		}
 
 #define SEQUENCE_DEBUG
@@ -839,6 +831,7 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 	char kind;
 	int status;
 	PreparedStatement *stmt;
+	int deadlock_detected = 0;
 
 	/* read Execute packet */
 	if (pool_read(frontend, &len, sizeof(len)) < 0)
@@ -851,51 +844,66 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 
 	stmt = lookup_prepared_statement_by_portal(&prepared_list, string);
 
-	/* load balance trick */
-	if (stmt && load_balance_enabled(backend, stmt->prepared_string))
-		start_load_balance(backend);
-	else if (MASTER_SLAVE)
+	if (stmt)
 	{
-		master_slave_was_enabled = 1;
-		MASTER_SLAVE = 0;
-		master_slave_dml = 1;
-	}
-	else if (REPLICATION && is_select_query(stmt->prepared_string) && !is_sequence_query(stmt->prepared_string))
-	{
-		int i;
-
-		/* save backend connection slots */
-		for (i=0;i<backend->num;i++)
+		/* load balance trick */
+		if (stmt && load_balance_enabled(backend, stmt->prepared_string))
+			start_load_balance(backend);
+		else if (MASTER_SLAVE)
 		{
-			slots[i] = backend->slots[i];
+			master_slave_was_enabled = 1;
+			MASTER_SLAVE = 0;
+			master_slave_dml = 1;
 		}
+		else if (REPLICATION && !pool_config.replicate_select && is_select_query(stmt->prepared_string) && !is_sequence_query(stmt->prepared_string))
+		{
+			int i;
 
-		/* send query to master only. */
-		replication_was_enabled = 1;
-		REPLICATION = 0;
-		in_load_balance = 1;
-		select_in_transaction = 1;
+			/* save backend connection slots */
+			for (i=0;i<backend->num;i++)
+			{
+				slots[i] = backend->slots[i];
+			}
+
+			/* send query to master only. */
+			replication_was_enabled = 1;
+			REPLICATION = 0;
+			in_load_balance = 1;
+			select_in_transaction = 1;
+		}
 	}
 
 	for (i = 0;i < backend->num;i++)
 	{
 		POOL_CONNECTION *cp = backend->slots[i]->con;
 
-		/* forward the query to the backend */
-		pool_write(cp, "E", 1);
-		sendlen = htonl(len + 4);
-		pool_write(cp, &sendlen, sizeof(sendlen));
-		pool_write(cp, string, len);
-
-		/*
-		 * send "Flush" message so that backend notices us
-		 * the completion of the command
-		 */
-		pool_write(cp, "H", 1);
-		sendlen = htonl(4);
-		if (pool_write_and_flush(cp, &sendlen, sizeof(sendlen)) < 0)
+		if (deadlock_detected)
 		{
-			return POOL_ERROR;
+			pool_write(cp, "Q", 1);
+			len = strlen(POOL_ERROR_QUERY) + 1;
+			sendlen = htonl(len + 4);
+			pool_write(cp, &sendlen, sizeof(sendlen));
+			pool_write_and_flush(cp, POOL_ERROR_QUERY, len);
+		}
+		else
+		{
+			/* forward the query to the backend */
+			pool_write(cp, "E", 1);
+			sendlen = htonl(len + 4);
+			pool_write(cp, &sendlen, sizeof(sendlen));
+			pool_write(cp, string, len);
+
+			/*
+			 * send "Flush" message so that backend notices us
+			 * the completion of the command
+			 */
+			pool_write(cp, "H", 1);
+			sendlen = htonl(4);
+
+			if (pool_write_and_flush(cp, &sendlen, sizeof(sendlen)) < 0)
+			{
+				return POOL_ERROR;
+			}
 		}
 
 		if (!REPLICATION)
@@ -905,13 +913,27 @@ static POOL_STATUS Execute(POOL_CONNECTION *frontend,
 			pool_debug("waiting for backend completing the query");
 			if (synchronize(cp))
 				return POOL_END;
+
+			/*
+			 * We must check deadlock error because a aborted transaction
+			 * by detecting deadlock isn't same on all nodes.
+			 * If a transaction is aborted on master node, pgpool send a
+			 * error query to another nodes.
+			 */
+			deadlock_detected = detect_deadlock_error(MASTER(backend), MAJOR(backend));
+			if (deadlock_detected < 0)
+				return POOL_END;
 		}
 	}
 
 	while ((kind = pool_read_kind(backend)),
-		   (kind != 'C' && kind != 'E' && kind != 'l' && kind != 's'))
+		   (kind != 'C' && kind != 'E' && kind != 'I' && kind != 's'))
 	{
-		if (kind < 0)
+		if (kind == -2) /* kind mismatch */
+		{
+			return error_kind_mismatch(frontend, backend, 0, 0);
+		}
+		else if (kind < 0)
 		{
 			pool_error("Execute: pool_read_kind error");
 			return POOL_ERROR;
@@ -1084,6 +1106,13 @@ static POOL_STATUS ReadyForQuery(POOL_CONNECTION *frontend,
 		master_slave_dml = 0;
 	}
 
+	if (force_replication)
+	{
+		force_replication = 0;
+		REPLICATION = replication_was_enabled;
+		replication_was_enabled = 0;
+	}
+
 #ifdef NOT_USED
 	return ProcessFrontendResponse(frontend, backend);
 #endif
@@ -1958,7 +1987,6 @@ static POOL_STATUS ProcessFrontendResponse(POOL_CONNECTION *frontend,
 					MASTER_SLAVE = 0;
 					master_slave_dml = 1;
 				}
-
 				status = SimpleForwardToBackend(fkind, frontend, backend);
 				if (pool_flush(MASTER(backend)))
 					status = POOL_ERROR;
@@ -2208,6 +2236,11 @@ static void process_reporting(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *b
 	status[i].desc = "stop replication mode on fatal error";
 	i++;
 
+	status[i].name = "replicate_select";
+	snprintf(status[i].value, MAXVALLEN, "%d", pool_config.replicate_select);
+	status[i].desc = "non 0 if SELECT statement is replicated";
+	i++;
+
 	status[i].name = "reset_query_list";
 	*(status[i].value) = '\0';
 	for (j=0;j<pool_config.num_reset_queries;j++)
@@ -2527,8 +2560,8 @@ POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_C
 
 	/*
 	 * Check if packet kind == 'C'(Command complete), '1'(Parse
-	 * complete), '3'(Close complete). If so, then register or unregister
-	 * pending prepared statement.
+	 * complete), '3'(Close complete). If so, then register or
+	 * unregister pending prepared statement.
 	 */
 	if ((kind == 'C' || kind == '1' || kind == '3') &&
 		pending_function &&	pending_prepared_stmt)
@@ -2538,8 +2571,15 @@ POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_C
 	else if (kind == 'C' && select_in_transaction)
 		select_in_transaction = 0;
 
-	pending_function = NULL;
-	pending_prepared_stmt = NULL;
+	/* 
+	 * Remove a pending function if a received message is not
+	 * NoticeResponse.
+	 */
+	if (kind != 'N')
+	{
+		pending_function = NULL;
+		pending_prepared_stmt = NULL;
+	}
 
 	status = pool_read(MASTER(backend), &len, sizeof(len));
 	if (status < 0)
@@ -2626,13 +2666,11 @@ POOL_STATUS SimpleForwardToFrontend(char kind, POOL_CONNECTION *frontend, POOL_C
 
 		if (select_in_transaction)
 		{
-			do_command(SECONDARY(backend), "send invalid query from pgpool to abort transaction.",
-					   PROTO_MAJOR_V3, 0);
-			pool_write(SECONDARY(backend), "S", 1);
-			res1 = htonl(4);
-			if (pool_write_and_flush(SECONDARY(backend), &res1, sizeof(res1)) < 0)
+			if (TSTATE(backend) != 'E')
 			{
-				return POOL_END;
+				in_load_balance = 0;
+				REPLICATION = 1;
+				do_error_command(SECONDARY(backend), PROTO_MAJOR_V3);
 			}
 			select_in_transaction = 0;
 		}
@@ -2736,8 +2774,13 @@ POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CO
 
 	len = ntohl(sendlen) - 4;
 
-	if (len <= 0)
+	if (len == 0)
 		return POOL_CONTINUE;
+	else if (len < 0)
+	{
+		pool_error("SimpleFowardToBackend: invalid message length");
+		return POOL_END;
+	}
 
 	p = pool_read2(frontend, len);
 	if (p == NULL)
@@ -2874,6 +2917,12 @@ POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CO
 				break;
 		}
 
+
+		/*
+		 * Describe message with a portal name receive two messages.
+		 * 1. ParameterDescription
+		 * 2. RowDescriptions or NoData
+		 */
 		if (kind == 'D' && *p == 'S')
 		{
 			kind1 = pool_read_kind(backend);
@@ -2886,14 +2935,25 @@ POOL_STATUS SimpleForwardToBackend(char kind, POOL_CONNECTION *frontend, POOL_CO
 			pool_flush(frontend);
 		}
 
-		kind1 = pool_read_kind(backend);
-		if (kind1 < 0)
+		for (;;)
 		{
-			pool_error("SimpleForwardToBackend: pool_read_kind error");
-			return POOL_ERROR;
+			kind1 = pool_read_kind(backend);
+			if (kind1 < 0)
+			{
+				pool_error("SimpleForwardToBackend: pool_read_kind error");
+				return POOL_ERROR;
+			}
+			SimpleForwardToFrontend(kind1, frontend, backend);
+			if (pool_flush(frontend) < 0)
+				return POOL_ERROR;
+
+			/*
+			 * If warning or log messages are received, we must read
+			 * one message from backend.
+			 */
+			if (kind1 != 'N') /* Notice Message */
+				break;
 		}
-		SimpleForwardToFrontend(kind1, frontend, backend);
-		return pool_flush(frontend);
 	}
 	else
 	{
@@ -3213,6 +3273,7 @@ static POOL_STATUS do_command(POOL_CONNECTION *backend, char *query, int protoMa
 	int status;
 	char kind;
 	char *string;
+	int deadlock_detected;
 
 	pool_debug("do_command: Query: %s", query);
 
@@ -3231,6 +3292,16 @@ static POOL_STATUS do_command(POOL_CONNECTION *backend, char *query, int protoMa
 		return POOL_END;
 	}
 
+	/*
+	 * We must check deadlock error because a aborted transaction
+	 * by detecting deadlock isn't same on all nodes.
+	 * If a transaction is aborted on master node, pgpool send a
+	 * error query to another nodes.
+	 */
+	deadlock_detected = detect_deadlock_error(backend, protoMajor);
+	if (deadlock_detected < 0)
+		return POOL_END;
+
 	/*
 	 * Expecting CompleteCommand
 	 */
@@ -3243,7 +3314,7 @@ static POOL_STATUS do_command(POOL_CONNECTION *backend, char *query, int protoMa
 
 	if (kind != 'C')
 	{
-		pool_error("do_command: backend does not successfully complete command %s status %c", query, kind);
+		pool_log("do_command: backend does not successfully complete command %s status %c", query, kind);
 
 	}
 
@@ -3303,6 +3374,68 @@ static POOL_STATUS do_command(POOL_CONNECTION *backend, char *query, int protoMa
 		backend->tstate = kind;
 	}
 
+	return deadlock_detected ? POOL_DEADLOCK : POOL_CONTINUE;
+}
+
+/*
+ * Send syntax error query to abort transaction.
+ * We need to sync transaction status in transaction block.
+ * SELECT query is sended to master only.
+ * If SELECT is error, we must abort transaction on other nodes.
+ */
+static POOL_STATUS do_error_command(POOL_CONNECTION *backend, int protoMajor)
+{
+	int len;
+	int status;
+	char kind;
+	char *string;
+	char *error_query = "send invalid query from pgpool to abort transaction";
+
+	/* send the query to the backend */
+	pool_write(backend, "Q", 1);
+	len = strlen(error_query)+1;
+
+	if (protoMajor == PROTO_MAJOR_V3)
+	{
+		int sendlen = htonl(len + 4);
+		pool_write(backend, &sendlen, sizeof(sendlen));
+	}
+
+	if (pool_write_and_flush(backend, error_query, len) < 0)
+	{
+		return POOL_END;
+	}
+
+	/*
+	 * Expecting CompleteCommand
+	 */
+	status = pool_read(backend, &kind, sizeof(kind));
+	if (status < 0)
+	{
+		pool_error("do_command: error while reading message kind");
+		return POOL_END;
+	}
+
+	/*
+	 * read ErrorResponse message
+	 */
+	if (protoMajor == PROTO_MAJOR_V3)
+	{
+		if (pool_read(backend, &len, sizeof(len)) < 0)
+			return POOL_END;
+		len = ntohl(len) - 4;
+		string = pool_read2(backend, len);
+		if (string == NULL)
+			return POOL_END;
+		pool_debug("command tag: %s", string);
+	}
+	else
+	{
+		string = pool_read_string(backend, &len, 0);
+		if (string == NULL)
+			return POOL_END;
+	}
+
 	return POOL_CONTINUE;
 }
 
@@ -3340,7 +3473,7 @@ static POOL_STATUS insert_lock(POOL_CONNECTION_POOL *backend, char *query)
 	char *table;
 	char qbuf[1024];
 	POOL_STATUS status;
-	int i;
+	int i, deadlock_detected = 0;
 
 	/* insert_lock can be used in V3 only */
 	if (MAJOR(backend) != PROTO_MAJOR_V3)
@@ -3377,7 +3510,13 @@ static POOL_STATUS insert_lock(POOL_CONNECTION_POOL *backend, char *query)
 	/* issue lock table command */
 	for (i = 0;i < backend->num;i++)
 	{
-		status = do_command(backend->slots[i]->con, qbuf, PROTO_MAJOR_V3, 0);
+		if (deadlock_detected)
+			status = do_command(backend->slots[i]->con, POOL_ERROR_QUERY, PROTO_MAJOR_V3, 0);
+		else
+			status = do_command(backend->slots[i]->con, qbuf, PROTO_MAJOR_V3, 0);
+
+		if (status == POOL_DEADLOCK)
+			deadlock_detected = 1;
 	}
 
 	return status;
@@ -3533,6 +3672,12 @@ static PreparedStatement *get_prepared_command_portal_and_statement(char *query)
 	/* skip data type list */
 	while (*query && *query != ')')
 		query++;
+
+	if (!*query)
+	{
+		pool_debug("get_prepared_command_portal_and_statement: could not get statement");
+		return NULL;
+	}
 	query++;
 
 	/* skip spaces */
@@ -3830,3 +3975,120 @@ static void query_ps_status(char *query, POOL_CONNECTION_POOL *backend)
 
 	set_ps_display(psbuf, false);
 }
+
+static POOL_STATUS error_kind_mismatch(POOL_CONNECTION *frontend, POOL_CONNECTION_POOL *backend, int kind, int kind1)
+{
+	int sts;
+
+	pool_error("pool_process_query: kind does not match between backends master(%c) secondary(%c)",
+			   kind, kind1);
+	pool_send_error_message(frontend, MAJOR(backend), "XX000", 
+							"kind mismatch between backends", "",
+							"check data consistency between master and secondary", __FILE__, __LINE__);
+
+	/* health check */
+	sts = health_check();
+	if (sts == -1)
+	{
+		notice_backend_error(1);
+		exit(1);
+	}
+	else if (sts == -2)
+	{
+		notice_backend_error(0);
+		exit(1);
+	}
+	
+	if (pool_config.replication_stop_on_mismatch)
+		return POOL_FATAL;
+	else
+		return POOL_ERROR;
+}
+
+static int detect_deadlock_error(POOL_CONNECTION *master, int major)
+{
+	int deadlock = 0;
+	char kind;
+	int readlen = 0, len;
+	char *buf;
+	char *p, *str;
+
+	if ((buf = malloc(1024)) == NULL)
+	{
+		pool_error("detect_deadlock_error: malloc failed");
+		return -1;
+	}
+
+	if (pool_read(master, &kind, sizeof(kind)))
+		return POOL_END;
+	readlen += sizeof(kind);
+	p = buf;
+	memcpy(p, &kind, sizeof(kind));
+	p += sizeof(kind);
+
+	if (kind == 'E') /* deadlock error? */
+	{
+		/* read actual query */
+		if (major == PROTO_MAJOR_V3)
+		{
+			char *error_code;
+			
+			if (pool_read(master, &len, sizeof(len)) < 0)
+				return POOL_END;
+			readlen += sizeof(len);
+			memcpy(p, &len, sizeof(len));
+			p += sizeof(len);
+			
+			len = ntohl(len) - 4;
+			str = malloc(len);
+			pool_read(master, str, len);
+			readlen += len;
+			if (readlen > 1024)
+			{
+				buf = realloc(buf, readlen);
+				if (buf == NULL)
+				{
+					pool_error("detect_deadlock_error: malloc failed");
+					return -1;
+				}
+			}
+			memcpy(p, str, len);
+
+			error_code = str;
+			while (*error_code)
+			{
+				if (*error_code == 'C')
+				{
+					if (strcmp(error_code+1, DEADLOCK_ERROR_CODE) == 0) /* deadlock error */
+					{
+						pool_debug("SimpleQuery: receive deadlock error from master node.");
+						deadlock = 1;
+					}
+					break;
+				}
+				else
+					error_code = error_code + strlen(error_code) + 1;
+			}
+			free(str);
+		}
+		else
+		{
+			str = pool_read_string(master, &len, 0);
+			readlen += len;
+			if (readlen > 1024)
+			{
+				buf = realloc(buf, readlen);
+				if (buf == NULL)
+				{
+					pool_error("detect_deadlock_error: malloc failed");
+					return -1;
+				}
+			}
+			memcpy(p, str, len);
+		}
+	}
+	if (pool_unread(master, buf, readlen) != 0)
+		deadlock = -1;
+	free(buf);
+	return deadlock;
+}
